/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Pieter Ghysels, Francois-Henry Rouet, Xiaoye S. Li.
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */
#ifndef ETREE_HPP
#define ETREE_HPP

#include <cassert>
#include <stack>

namespace strumpack {

  template<typename integer_t> inline integer_t
  make_set(integer_t i, std::vector<integer_t>& pp) {
    pp[i] = i;
    return i;
  }

  template<typename integer_t> inline integer_t
  link(integer_t s, integer_t t, std::vector<integer_t>& pp) {
    pp[s] = t;
    return t;
  }

  /** path halving */
  template<typename integer_t> inline integer_t
  find(integer_t i, std::vector<integer_t>& pp) {
    integer_t p = pp[i];
    integer_t gp = pp[p];
    while (gp != p) {
      pp[i] = gp;
      i = gp;
      p = pp[i];
      gp = pp[p];
    }
    return p;
  }

  /*! \brief Symmetric elimination tree
   *
   * <pre>
   *      p = spsymetree (A);
   *
   *      Find the elimination tree for symmetric matrix A.
   *      This uses Liu's algorithm, and runs in time O(nz*log n).
   *
   *      Input:
   *        Square sparse matrix A.  No check is made for symmetry;
   *        elements below and on the diagonal are ignored.
   *        Numeric values are ignored, so any explicit zeros are
   *        treated as nonzero.
   *      Output:
   *        Integer array of parents representing the etree, with n
   *        meaning a root of the elimination forest.
   *      Note:
   *        This routine uses only the upper triangle, while sparse
   *        Cholesky (as in spchol.c) uses only the lower.  Matlab's
   *        dense Cholesky uses only the upper.  This routine could
   *        be modified to use the lower triangle either by transposing
   *        the matrix or by traversing it by rows with auxiliary
   *        pointer and link arrays.
   *
   *      John R. Gilbert, Xerox, 10 Dec 1990
   *      Based on code by JRG dated 1987, 1988, and 1990.
   *      Modified by X.S. Li, November 1999.
   * </pre>
   */
  template<typename integer_t> std::vector<integer_t>
  sp_symetree_dist(integer_t *acolst, integer_t *acolend, // column starts and ends past 1
		   integer_t *arow,                       // row indices of A
		   integer_t n,                           // dimension of A
		   integer_t subgraph_begin=0) {          // first row/column of subgraph
    // if working on subgraph, acolst/end only for subgraph and n is number of vertices in the subgraph
    std::vector<integer_t> root(n, 0); // root of subtee of etree
    std::vector<integer_t> pp(n, 0);
    std::vector<integer_t> parent(n);
    integer_t rset, cset, rroot, row, col, p;
    for (col=0; col<n; col++) {
      cset = make_set(col, pp);
      root[cset] = col;
      parent[col] = n;
      for (p=acolst[col]; p<acolend[col]; p++) {
	row = arow[p] - subgraph_begin;
	if (row >= col) continue;
	rset = find(row, pp);
	rroot = root[rset];
	if (rroot != col) {
	  parent[rroot] = col;
	  cset = link(cset, rset, pp);
	  root[cset] = col;
	}
      }
    }
    return parent;
  }

  /** Depth-first search from vertex n on the etree. Non-recursive version. */
  template<typename integer_t> void
  etdfs_non_recursive(std::vector<integer_t>& parent,
		      std::vector<integer_t>& first_kid, std::vector<integer_t>& next_kid,
		      std::vector<integer_t>& post, integer_t postnum) {
    integer_t n = parent.size();
    integer_t current = n, first, next;
    while (postnum != n) {
      first = first_kid[current];     // no kid for the current node
      if (first == -1) {              // no first kid for the current node
	post[current] = postnum++;    // numbering this node because it has no kid
	next = next_kid[current];     // looking for the next kid
	while (next == -1) {
	  current = parent[current];  // no more kids : back to the parent node
	  post[current] = postnum++;  // numbering the parent node
	  next = next_kid[current];   // get the next kid
	}
	if (postnum == n+1) return;   // stopping criterion
	current = next;               // updating current node
      } else current = first;         // updating current node
    }
  }

  /** Post order an etree. */
  template<typename integer_t> std::vector<integer_t>
  etree_postorder(std::vector<integer_t>& parent) {
    auto n = parent.size();
    std::vector<integer_t> first_kid(n+1, integer_t(-1));
    std::vector<integer_t> next_kid(n+1);
    // set up structure describing children
    for (integer_t v = n-1; v >= 0; v--) {
      integer_t dad = parent[v];
      next_kid[v] = first_kid[dad];
      first_kid[dad] = v;
    }
    // depth-first search from dummy root vertex #n
    std::vector<integer_t> post(n+1);
    etdfs_non_recursive(parent, first_kid, next_kid, post, integer_t(0));
    return post;
  }

  template<typename integer_t> integer_t etree_to_separator_tree
  (std::vector<integer_t>& etree, std::vector<integer_t>& parent,
   std::vector<integer_t>& lchild, std::vector<integer_t>& rchild,
   std::vector<integer_t>& level, std::vector<integer_t>& sep_ptr) {
    integer_t n = etree.size();
    if (n == 0) return 0;
    // in etree a root has parent n, replace n with -1
    std::replace(etree.begin(), etree.end(), n, integer_t(-1));
    integer_t nr_roots = std::count(etree.begin(), etree.end(), integer_t(-1));
    // deal with multiple roots
    if (nr_roots > 1) {
      for (integer_t r=0; r<nr_roots-1; r++) {
	auto hi = etree.size() - 1;
	while (etree[hi] != -1) hi--;
	auto root_right = hi;
	hi--;
	while (etree[hi] != -1) hi--;
	integer_t max_p = etree.size();
	etree.push_back(-1);
	etree[root_right] = max_p;
	etree[hi] = max_p;
      }
    }
    auto new_n = etree.size();
    std::vector<integer_t> count(new_n, 0);
    std::vector<integer_t> etree_lchild(new_n, integer_t(-1));
    std::vector<integer_t> etree_rchild(new_n, integer_t(-1));
    for (size_t i=0; i<new_n; i++) {
      integer_t p = etree[i];
      if (p != -1) {
	count[p]++;
	switch (count[p]) {
	case 1: etree_lchild[p] = i; break; // node i is the first child of node p
	case 2: etree_rchild[p] = i; break; // node i is the second child of node p
	case 3: // node i is the third child of node p
	  // make a new node with children the first two children of p,
	  //     set this new node as the left child of p
	  // make node i the right child of p
	  integer_t max_p = etree.size();
	  etree.push_back(max_p);
	  etree_lchild.push_back(etree_lchild[p]);
	  etree_rchild.push_back(etree_rchild[p]);
	  etree_lchild[p] = max_p;
	  etree_rchild[p] = i;
	  count[p]--;
	  break;
	}
      }
    }
    sep_ptr.push_back(0);
    level.clear();
    integer_t nr_sep = 0;
    integer_t root_id = std::distance(etree.begin(), std::find(etree.begin(), etree.end(), integer_t(-1)));
    std::stack<integer_t,std::vector<integer_t> > s;
    std::stack<integer_t,std::vector<integer_t> > l;
    s.push(root_id);
    integer_t prev=-1;
    while (!s.empty()) {
      integer_t i = s.top();
      if (prev == -1 || etree_lchild[prev] == i || etree_rchild[prev] == i) { // moving down
    	if      (etree_lchild[i] != -1) s.push(etree_lchild[i]);     // go down left
    	else if (etree_rchild[i] != -1) s.push(etree_rchild[i]);     // if no left, then go down right
      } else if (etree_lchild[i] == prev) {                          // moving up from the left,
    	if (etree_rchild[i] != -1) {
	  l.push(parent.size() - 1);
	  s.push(etree_rchild[i]);          // go down right
	}
      } else {
    	// skip nodes that have only one child, this will group nodes in fronts
    	if ((etree_rchild[i] == -1 && etree_lchild[i] == -1) ||  // two children
    	    (etree_rchild[i] != -1 && etree_lchild[i] != -1)) {  // or no children
    	  integer_t pid = parent.size();
    	  parent.push_back(-1);
    	  if (etree_lchild[i] != -1) {
    	    lchild.push_back(l.top());
    	    parent[l.top()] = pid;
	    l.pop();
    	  } else lchild.push_back(-1);
    	  if (etree_rchild[i] != -1) {
    	    rchild.push_back(pid - 1);
    	    parent[pid-1] = pid;
    	  } else rchild.push_back(-1);
    	  sep_ptr.push_back(sep_ptr.back());
    	  nr_sep++;
    	}
    	// nodes >= n are empty separators introduced to avoid nodes with three children
    	// so do not count those when computing separator size!
    	if (i < n) sep_ptr.back()++;
	s.pop();
      }
      prev = i;
    }

    level.resize(nr_sep, 0);
    std::function<void(integer_t,integer_t)> set_level = [&](integer_t i, integer_t l) {
      level[i] = l;
      if (lchild[i] != -1) set_level(lchild[i], l+1);
      if (rchild[i] != -1) set_level(rchild[i], l+1);
    };
    set_level(nr_sep-1, 0);
    integer_t max_level = *std::max_element(level.begin(), level.end());
    std::for_each(level.begin(), level.end(), [&](integer_t& l){ l=max_level-l+1; });
    return nr_sep;
  }

} // end namespace strumpack

#endif // ETREE_HPP
