/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Pieter Ghysels, Francois-Henry Rouet, Xiaoye S. Li.
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */
#ifndef ELIMINATION_TREE_MPI_HPP
#define ELIMINATION_TREE_MPI_HPP

#include <iostream>
#include <algorithm>
#include <random>
#include <algorithm>
#include "blas_lapack_wrapper.hpp"
#include "CompressedSparseMatrix.hpp"
#include "CSRMatrixMPI.hpp"
#include "MatrixReorderingMPI.hpp"
#include "FrontalMatrix.hpp"
#include "FrontalMatrixMPI.hpp"
#include "FrontalMatrixDenseMPI.hpp"
#include "FrontalMatrixHSSMPI.hpp"
#include "strumpack_parameters.hpp"

namespace strumpack {

  template<typename scalar_t,typename real_t,typename integer_t>
  class EliminationTreeMPI : public EliminationTree<scalar_t,real_t,integer_t> {
  public:
    EliminationTreeMPI(MPI_Comm _comm);
    EliminationTreeMPI(CompressedSparseMatrix<scalar_t,real_t,integer_t>* A,
		       MatrixReordering<scalar_t,real_t,integer_t>* nd,
		       bool use_hss, MPI_Comm _comm);
    virtual ~EliminationTreeMPI() { mpi_free_comm(&comm); }
    void multifrontal_solve(scalar_t* x);
    integer_t get_maximum_rank();
    long long get_factor_nonzeros();
    long long get_dense_factor_nonzeros();
    using SepRange = std::pair<integer_t,integer_t>;
    std::vector<SepRange> subtree_ranges;

  protected:
    using F_t = FrontalMatrix<scalar_t,real_t,integer_t>;
    using FD_t = FrontalMatrixDense<scalar_t,real_t,integer_t>;
    using FHSS_t = FrontalMatrixHSS<scalar_t,real_t,integer_t>;
    using FMPI_t = FrontalMatrixMPI<scalar_t,real_t,integer_t>;
    using FDMPI_t = FrontalMatrixDenseMPI<scalar_t,real_t,integer_t>;
    using FHSSMPI_t = FrontalMatrixHSSMPI<scalar_t,real_t,integer_t>;

    MPI_Comm comm;
    int rank;

    virtual int get_nr_HSS_fronts() {
      MPI_Allreduce(MPI_IN_PLACE, &this->nr_HSS_fronts, 1, MPI_INT, MPI_SUM, comm);
      return this->nr_HSS_fronts;
    }
    virtual int get_nr_dense_fronts() {
      MPI_Allreduce(MPI_IN_PLACE, &this->nr_dense_fronts, 1, MPI_INT, MPI_SUM, comm);
      return this->nr_dense_fronts;
    }

  private:
    CompressedSparseMatrix<scalar_t,real_t,integer_t>* A = NULL;
    MatrixReordering<scalar_t,real_t,integer_t>* nd = NULL;

    struct ParFront {
      // TODO store a pointer to the actual front??
      ParFront(integer_t _sep_begin, integer_t _dim_sep, int _P0, int _P, int _ctxt, int _ctxt_all)
	: sep_begin(_sep_begin), dim_sep(_dim_sep), P0(_P0), P(_P), ctxt(_ctxt), ctxt_all(_ctxt_all) {}
      integer_t sep_begin; integer_t dim_sep; int P0; int P; int ctxt; int ctxt_all;
    };
    std::vector<ParFront> parallel_fronts;
    integer_t active_pfronts;

    F_t* proportional_mapping(std::vector<integer_t>* upd, float* subtree_work, SepRange& local_range,
			      integer_t sep, int P0, int P, MPI_Comm front_comm,
			      bool keep, bool is_hss, bool use_hss, int level=0);
    void symbolic_factorization(integer_t sep, std::vector<integer_t>* upd, float* subtree_work, int depth=0);

    void log_etree(std::vector<integer_t>* upd, float* subtree_work);

    void sequential_to_block_cyclic(scalar_t* x, DistributedMatrix<scalar_t,real_t>*& x_dist);
    void block_cyclic_to_sequential(scalar_t* x, DistributedMatrix<scalar_t,real_t>*& x_dist);
  };

  template<typename scalar_t,typename real_t,typename integer_t>
  EliminationTreeMPI<scalar_t,real_t,integer_t>::EliminationTreeMPI(MPI_Comm _comm)
    : EliminationTree<scalar_t,real_t,integer_t>() {
    MPI_Comm_dup(_comm, &comm);
  }

  template<typename scalar_t,typename real_t,typename integer_t>
  EliminationTreeMPI<scalar_t,real_t,integer_t>::EliminationTreeMPI
  (CompressedSparseMatrix<scalar_t,real_t,integer_t>* _A, MatrixReordering<scalar_t,real_t,integer_t>* _nd,
   bool use_hss, MPI_Comm _comm) : EliminationTree<scalar_t,real_t,integer_t>(),
    A(_A), nd(_nd), active_pfronts(0) {
    MPI_Comm_dup(_comm, &comm);
    rank = mpi_rank(comm);
    auto P = mpi_nprocs(comm);
    auto upd = new std::vector<integer_t>[nd->nbsep];
    auto subtree_work = new float[nd->nbsep];
#pragma omp parallel
#pragma omp single
    {
      symbolic_factorization(nd->nbsep-1, upd, subtree_work);
    }
    SepRange local_range{A->size(), 0};

    MPI_Comm tree_comm;
    if (P>1) MPI_Comm_dup(comm, &tree_comm);
    else tree_comm = comm;
    this->etree_root = proportional_mapping(upd, subtree_work, local_range, nd->nbsep-1,
					    0, P, tree_comm, true, true, use_hss, 0);
    subtree_ranges.resize(P);
    MPI_Allgather(&local_range, sizeof(SepRange), MPI_BYTE, subtree_ranges.data(),
     		  sizeof(SepRange), MPI_BYTE, comm);
    if (params::log_etree) log_etree(upd, subtree_work);
    delete[] upd;
    delete[] subtree_work;
    nd->clear_tree_data();
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  EliminationTreeMPI<scalar_t,real_t,integer_t>::log_etree
  (std::vector<integer_t>* upd, float* subtree_work) {
    if (rank) return;
    std::ofstream log_file("etree.log", std::ios::out);
    log_file << "PERMUTATION" << std::endl;
    log_file << "order = [ ";
    for (size_t i=0; i<nd->order.size(); i++) log_file << nd->order[i] << " ";
    log_file << "]" << std::endl << "iorder = [ ";
    for (size_t i=0; i<nd->iorder.size(); i++) log_file << nd->iorder[i] << " ";
    log_file << "]" << std::endl << std::endl;
    log_file << "Symbolic factorization" << std::endl;
    for (integer_t sep=0; sep<nd->nbsep; sep++) {
      auto chl = nd->left_child[sep];
      auto chr = nd->right_child[sep];
      auto sep_begin = nd->sep_ptr[sep];
      auto sep_end = nd->sep_ptr[sep+1];
      log_file << "sep=" << sep << ", lch=" << chl << ", rch=" << chr << ", pa=" << nd->parent[sep]
	       << ", lvl=" << nd->level[sep] << ", work=" << subtree_work[sep]
	       << ", sep_begin=" << sep_begin << ", sep_end=" << sep_end;
      log_file << ", upd= [ ";
      for (size_t i=0; i<upd[sep].size(); i++) log_file << upd[sep][i] << " ";
      log_file << "]" << std::endl;
    }
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  EliminationTreeMPI<scalar_t,real_t,integer_t>::sequential_to_block_cyclic
  (scalar_t* x, DistributedMatrix<scalar_t,real_t>*& x_dist) {
    size_t pos = 0;
    for (auto& pf : parallel_fronts)
      if (rank >= pf.P0 && rank < pf.P0+pf.P) pos++;
    x_dist = new DistributedMatrix<scalar_t,real_t>[pos];
    pos = 0;
    for (auto& pf : parallel_fronts)
      if (rank >= pf.P0 && rank < pf.P0+pf.P)
	// TODO this also does a pgemr2d!
	x_dist[pos++] = DistributedMatrix<scalar_t,real_t>
	  (pf.ctxt, pf.dim_sep, 1, x + pf.sep_begin, pf.ctxt_all);
  }

  // TODO: rewrite this with a single alltoallv/allgatherv
  template<typename scalar_t,typename real_t,typename integer_t> void
  EliminationTreeMPI<scalar_t,real_t,integer_t>::block_cyclic_to_sequential
  (scalar_t* x, DistributedMatrix<scalar_t,real_t>*& x_dist) {
    auto P = mpi_nprocs(comm);
    auto cnts = new int[2*P];
    auto disp = cnts + P;
    for (int p=0; p<P; p++) {
      cnts[p] = std::max(integer_t(0), subtree_ranges[p].second - subtree_ranges[p].first);
      disp[p] = subtree_ranges[p].first;
    }
    MPI_Allgatherv(MPI_IN_PLACE, 0, mpi_type<scalar_t>(), x, cnts, disp, mpi_type<scalar_t>(), comm);
    delete[] cnts;

    auto xd = x_dist;
    for (auto& pf : parallel_fronts) {
      if (rank >= pf.P0 && rank < pf.P0+pf.P)
	(xd++)->get_sequential(x+pf.sep_begin, pf.ctxt_all);
      MPI_Bcast(x+pf.sep_begin, pf.dim_sep, mpi_type<scalar_t>(), pf.P0, comm);
    }
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  EliminationTreeMPI<scalar_t,real_t,integer_t>::multifrontal_solve(scalar_t* x) {
    DistributedMatrix<scalar_t,real_t>* x_dist;
    sequential_to_block_cyclic(x, x_dist);
    this->etree_root->multifrontal_solve(x, x_dist, this->wmem);
    block_cyclic_to_sequential(x, x_dist);
    delete[] x_dist;
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  EliminationTreeMPI<scalar_t,real_t,integer_t>::symbolic_factorization
  (integer_t sep, std::vector<integer_t>* upd, float* subtree_work, int depth) {
    auto chl = nd->left_child[sep];
    auto chr = nd->right_child[sep];
    if (depth < params::task_recursion_cutoff_level) {
      if (chl != -1)
#pragma omp task untied default(shared) final(depth >= params::task_recursion_cutoff_level-1) mergeable
	symbolic_factorization(chl, upd, subtree_work, depth+1);
      if (chr != -1)
#pragma omp task untied default(shared) final(depth >= params::task_recursion_cutoff_level-1) mergeable
	symbolic_factorization(chr, upd, subtree_work, depth+1);
#pragma omp taskwait
    } else {
      if (chl != -1) symbolic_factorization(chl, upd, subtree_work, depth);
      if (chr != -1) symbolic_factorization(chr, upd, subtree_work, depth);
    }
    auto sep_begin = nd->sep_ptr[sep];
    auto sep_end = nd->sep_ptr[sep+1];
    if (sep != nd->nbsep-1) { // not necessary for the root
      for (integer_t c=sep_begin; c<sep_end; c++) {
	auto ice = A->get_ind()+A->get_ptr()[c+1];
	auto icb = std::lower_bound(A->get_ind()+A->get_ptr()[c], ice, sep_end);
	auto mid = upd[sep].size();
	std::copy(icb, ice, std::back_inserter(upd[sep]));
	std::inplace_merge(upd[sep].begin(), upd[sep].begin() + mid, upd[sep].end());
	upd[sep].erase(std::unique(upd[sep].begin(), upd[sep].end()), upd[sep].end());
      }
      if (chl != -1) {
	auto icb = std::lower_bound(upd[chl].begin(), upd[chl].end(), sep_end);
	auto mid = upd[sep].size();
	std::copy(icb, upd[chl].end(), std::back_inserter(upd[sep]));
	std::inplace_merge(upd[sep].begin(), upd[sep].begin() + mid, upd[sep].end());
	upd[sep].erase(std::unique(upd[sep].begin(), upd[sep].end()), upd[sep].end());
      }
      if (chr != -1) {
	auto icb = std::lower_bound(upd[chr].begin(), upd[chr].end(), sep_end);
	auto mid = upd[sep].size();
	std::copy(icb, upd[chr].end(), std::back_inserter(upd[sep]));
	std::inplace_merge(upd[sep].begin(), upd[sep].begin() + mid, upd[sep].end());
	upd[sep].erase(std::unique(upd[sep].begin(), upd[sep].end()), upd[sep].end());
      }
    }
    integer_t dim_blk = (sep_end - sep_begin) + upd[sep].size();
    // assume amount of work per front is N^3, work per subtree is work on front plus children
    float wl = (chl != -1) ? subtree_work[chl] : 0.;
    float wr = (chr != -1) ? subtree_work[chr] : 0.;
    subtree_work[sep] = (float(dim_blk)*dim_blk*dim_blk) + wl + wr;
  }

  // keep track of [P0_pa, P0_pa+P_pa) -> can be used to stop iso keep_subtree
  template<typename scalar_t,typename real_t,typename integer_t> FrontalMatrix<scalar_t,real_t,integer_t>*
  EliminationTreeMPI<scalar_t,real_t,integer_t>::proportional_mapping
  (std::vector<integer_t>* upd, float* subtree_work, SepRange& local_range, integer_t sep,
   int P0, int P, MPI_Comm front_comm, bool keep, bool hss_parent, bool use_hss, int level) {
    auto sep_begin = nd->sep_ptr[sep];
    auto sep_end = nd->sep_ptr[sep+1];
    auto dim_sep = sep_end - sep_begin;
    F_t* front = NULL;
    bool is_hss = use_hss && hss_parent && (dim_sep >= params::hss_minimum_front_size);
    if (rank == P0) {
      if (is_hss) this->nr_HSS_fronts++;
      else this->nr_dense_fronts++;
    }
    if (P == 1) {
      if (keep) {
	if (is_hss) {
	  front = new FHSS_t(A, sep, sep_begin, sep_end, upd[sep].size(), upd[sep].data());
	  front->set_HSS_partitioning(nd->sep_HSS_tree[sep], level == 0);
	} else front = new FD_t(A, sep, sep_begin, sep_end, upd[sep].size(), upd[sep].data());
      }
      if (P0 == rank) {
	local_range.first  = std::min(local_range.first,  sep_begin);
	local_range.second = std::max(local_range.second, sep_end);
      }
    } else {
      if (keep) {
	if (is_hss) {
	  front = new FHSSMPI_t(A, active_pfronts, sep_begin, sep_end, upd[sep].size(), upd[sep].data(), front_comm, P);
	  front->set_HSS_partitioning(nd->sep_HSS_tree[sep], level == 0);
	} else front = new FDMPI_t(A, active_pfronts, sep_begin, sep_end, upd[sep].size(), upd[sep].data(), front_comm, P);
	if (rank >= P0 && rank < P0+P) active_pfronts++;
      } else mpi_free_comm(&front_comm);
      int ctxt = front ? static_cast<FMPI_t*>(front)->blacs_context() : -1;
      int ctxt_all = front ? static_cast<FMPI_t*>(front)->blacs_context_all() : -1;
      parallel_fronts.emplace_back(sep_begin, sep_end-sep_begin, P0, P, ctxt, ctxt_all);
    }

    // only store a node if you are part of its communicator
    // and also store your siblings!! needed for extend-add
    if (rank < P0 || rank >= P0+P) keep = false;
    if (P == 1 && P0 != rank) return front;

    auto chl = nd->left_child[sep];
    auto chr = nd->right_child[sep];
    if (chl != -1) {
      float wl = subtree_work[chl];
      float wr = (chr != -1) ? subtree_work[chr] : 0.;
      int Pl = std::max(1, std::min(int(std::round(P * wl / (wl + wr))), P-1));
      int Pr = std::max(1, P - Pl);
      auto fl = proportional_mapping(upd, subtree_work, local_range, chl, P0, Pl,
				     mpi_sub_comm(front_comm, 0, Pl),
				     keep, is_hss, use_hss, level+1);
      if (front) front->lchild = fl;
      if (chr != -1) {
	auto fr = proportional_mapping(upd, subtree_work, local_range, chr, P0+P-Pr, Pr,
				       mpi_sub_comm(front_comm, P-Pr, Pr),
				       keep, is_hss, use_hss, level+1);
	if (front) front->rchild = fr;
      }
    } else {
      if (chr != -1) {
	auto fr = proportional_mapping(upd, subtree_work, local_range, chr, P0, P,
				       front_comm, keep, is_hss, use_hss, level+1);
	if (front) front->rchild = fr;
      }
    }
    return front;
  }

  template<typename scalar_t,typename real_t,typename integer_t> integer_t
  EliminationTreeMPI<scalar_t,real_t,integer_t>::get_maximum_rank() {
    integer_t max_rank = EliminationTree<scalar_t,real_t,integer_t>::get_maximum_rank();
    MPI_Allreduce(MPI_IN_PLACE, &max_rank, 1, mpi_type<integer_t>(), MPI_MAX, comm);
    return max_rank;
  }

  template<typename scalar_t,typename real_t,typename integer_t> long long
  EliminationTreeMPI<scalar_t,real_t,integer_t>::get_factor_nonzeros() {
    long long nonzeros = EliminationTree<scalar_t,real_t,integer_t>::get_factor_nonzeros();
    MPI_Allreduce(MPI_IN_PLACE, &nonzeros, 1, MPI_LONG_LONG_INT, MPI_SUM, comm);
    return nonzeros;
  }

  template<typename scalar_t,typename real_t,typename integer_t> long long
  EliminationTreeMPI<scalar_t,real_t,integer_t>::get_dense_factor_nonzeros() {
    long long nonzeros = EliminationTree<scalar_t,real_t,integer_t>::get_dense_factor_nonzeros();
    MPI_Allreduce(MPI_IN_PLACE, &nonzeros, 1, MPI_LONG_LONG_INT, MPI_SUM, comm);
    return nonzeros;
  }

} // end namespace strumpack

#endif
