/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Pieter Ghysels, Francois-Henry Rouet, Xiaoye S. Li.
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */
#ifndef MATRIX_REORDERING_MPI_HPP
#define MATRIX_REORDERING_MPI_HPP
#include <unordered_map>
#include <algorithm>
#include "CSRGraph.hpp"
#include "parmetis.h"

namespace strumpack {

  inline Partition_tree* HSSPartitionTree_to_Partition_tree(HSSPartitionTree& in_tree) {
    auto out_tree = new Partition_tree();
    out_tree->size = in_tree.size;
    assert(in_tree.c.size() == 0 || in_tree.c.size() == 2);
    for (size_t c=0; c<in_tree.c.size(); c++)
      out_tree->c[c] = HSSPartitionTree_to_Partition_tree(in_tree.c[c]);
    assert(out_tree->check() == true);
    return out_tree;
  }

  inline void destroy_Partition_tree(Partition_tree* tree) {
    if (tree) {
      destroy_Partition_tree(tree->c[0]);
      destroy_Partition_tree(tree->c[1]);
      delete tree;
    }
  }

  template<typename scalar_t,typename real_t,typename integer_t>
  class MatrixReorderingMPI
    : public MatrixReordering<scalar_t,real_t,integer_t> {
  public:
    MatrixReorderingMPI(MPI_Comm c) : my_sub_graph(NULL), my_dist_sep(NULL), comm(c) {}
    virtual ~MatrixReorderingMPI() { delete my_sub_graph; delete my_dist_sep; }
    virtual void separator_reordering(CompressedSparseMatrix<scalar_t,real_t,integer_t>* A, bool verbose);
    virtual void clear_tree_data();

    integer_t dist_nr_sep;
    std::vector<integer_t> dist_lchild;
    std::vector<integer_t> dist_rchild;
    std::vector<integer_t> dist_parent;
    std::vector<integer_t> dist_sizes;
    std::vector<integer_t> dist_sep_ptr;
    std::vector<integer_t> dist_level;
    std::unordered_map<integer_t,HSSPartitionTree> dsep_HSS_tree;

    /**
     * proc_dist_sep[sep] holds the rank of the process responsible
     * for distributed separator sep
     * - if sep is a leaf of the distributed tree, proc_dist_sep
     *    points to the process holding that subgraph as my_sub_graph
     * - if sep is a non-leaf, proc_dist_sep points to the process
     *    holding the graph of the distributed separator sep as
     *    my_dist_sep
     */
    std::vector<integer_t> proc_dist_sep;

    /**
     * Every process is responsible for one local subgraph of A.  The
     * distributed nested dissection will create a separator tree with
     * exactly P leafs.  Each process takes one of those leafs and
     * stores the corresponding part of the permuted matrix A in
     * sub_graph_A.
     */
    CSRGraph<integer_t>* my_sub_graph;
    std::pair<integer_t,integer_t> sub_graph_range;
    std::vector<std::pair<integer_t,integer_t>> sub_graph_ranges;

    /**
     * Every process is responsible for one separator from the
     * distributed part of nested dissection.  The graph of the
     * distributed separator for which this rank is responsible is
     * stored as dist_sep_A.
     */
    CSRGraph<integer_t>* my_dist_sep;
    std::pair<integer_t,integer_t> dist_sep_range;
    std::vector<std::pair<integer_t,integer_t>> dist_sep_ranges;

  protected:
    MPI_Comm comm;
    void get_local_graphs(CSRMatrixMPI<scalar_t,real_t,integer_t>* A_mpi);
    void build_sep_tree_from_order(CSRMatrixMPI<scalar_t,real_t,integer_t>* A_mpi);
    void allgather_local_order(CSRMatrixMPI<scalar_t,real_t,integer_t>* A_mpi, integer_t* local_order);
    void distributed_separator_reordering(integer_t* global_sep_order);
    void local_separator_reordering(integer_t* global_sep_order);
    virtual void nested_dissection_print(integer_t n, integer_t nnz, int stratnum, int stratpar, bool verbose);
  };

  template<typename scalar_t,typename real_t,typename integer_t> void
  MatrixReorderingMPI<scalar_t,real_t,integer_t>::separator_reordering
  (CompressedSparseMatrix<scalar_t,real_t,integer_t>* A, bool verbose) {
    auto n = A->size();
    auto global_sep_order = new integer_t[n];
    for (integer_t i=0; i<n; i++) global_sep_order[i] = i;

    distributed_separator_reordering(global_sep_order);
    local_separator_reordering(global_sep_order);

    for (integer_t i=0; i<n; i++) this->order[global_sep_order[i]] = this->iorder[i];
    for (integer_t i=0; i<n; i++) this->iorder[this->order[i]] = i;
    std::swap(this->order, this->iorder);
    delete[] global_sep_order;
  }

  template<typename integer_t> inline int WRAPPER_METIS_PartGraphRecursive
  (idx_t* nvtxs, idx_t* ncon, std::vector<integer_t>& sep_csr_ptr, std::vector<integer_t>& sep_csr_ind,
   idx_t* two, idx_t* edge_cut, idx_t* partitioning) {
    auto ptr = new idx_t[sep_csr_ptr.size()+sep_csr_ind.size()];
    auto ind = ptr + sep_csr_ptr.size();
    for (size_t i=0; i<sep_csr_ptr.size(); i++) ptr[i] = sep_csr_ptr[i];
    for (size_t i=0; i<sep_csr_ind.size(); i++) ind[i] = sep_csr_ind[i];
    int ierr = METIS_PartGraphRecursive(nvtxs, ncon, ptr, ind, NULL, NULL, NULL,
					two, NULL, NULL, NULL, edge_cut, partitioning);
    delete[] ptr;
    return ierr;
  }
  template<> inline int WRAPPER_METIS_PartGraphRecursive
  (idx_t* nvtxs, idx_t* ncon, std::vector<idx_t>& sep_csr_ptr, std::vector<idx_t>& sep_csr_ind,
   idx_t* two, idx_t* edge_cut, idx_t* partitioning) {
    return METIS_PartGraphRecursive(nvtxs, ncon, sep_csr_ptr.data(), sep_csr_ind.data(),
				    NULL, NULL, NULL, two, NULL, NULL, NULL, edge_cut, partitioning);
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  MatrixReorderingMPI<scalar_t,real_t,integer_t>::local_separator_reordering(integer_t* global_sep_order) {
    auto n_local = my_sub_graph->size();
    auto local_sep_order = new integer_t[2*n_local];
    auto local_sep_iorder = local_sep_order + n_local;
    for (integer_t i=0; i<n_local; i++) local_sep_order[i] = i + sub_graph_range.first;
    std::vector<integer_t> sep_csr_ptr, sep_csr_ind;

    std::function<void(integer_t)> reorder_separator = [&](integer_t sep) {
      auto sep_begin = this->sep_ptr[sep];
      auto sep_end = this->sep_ptr[sep+1];
      auto dim_sep = sep_end - sep_begin;
      if (dim_sep < params::hss_minimum_front_size) return; // this front (and its descendants) are not HSS
      HSSPartitionTree tree(dim_sep);
      if (dim_sep > 2 * params::hss_minimum_partition_size) {
	std::function<void(HSSPartitionTree&, integer_t&, integer_t, integer_t)> split =
	  [&](HSSPartitionTree& hss_tree, integer_t& nr_parts, integer_t part, integer_t count) {
	  my_sub_graph->extract_separator_subgraph(sub_graph_range.first, sep_begin, sep_end,
						   part, local_sep_order+sep_begin,
						   sep_csr_ptr, sep_csr_ind);
	  idx_t ncon = 1, edge_cut = 0, two = 2, nvtxs = sep_csr_ptr.size()-1;
	  auto partitioning = new idx_t[nvtxs];
	  int info = WRAPPER_METIS_PartGraphRecursive(&nvtxs, &ncon, sep_csr_ptr, sep_csr_ind,
						      &two, &edge_cut, partitioning);
	  if (info !=  METIS_OK) {
	    std::cerr << "METIS_PartGraphRecursive for separator reordering returned: " << info << std::endl;
	    exit(1);
	  }
	  hss_tree.c.resize(2);
	  for (integer_t i=sep_begin, j=0; i<sep_end; i++)
	    if (local_sep_order[i] == part) {
	      auto p = partitioning[j++];
	      local_sep_order[i] = -count - p;
	      hss_tree.c[p].size++;
	    }
	  for (integer_t p=0; p<2; p++)
	    if (hss_tree.c[p].size > 2 * params::hss_minimum_partition_size)
	      split(hss_tree.c[p], nr_parts, -count-p, count+2);
	    else std::replace(local_sep_order+sep_begin, local_sep_order+sep_end, -count-p, nr_parts++);
	  delete[] partitioning;
	};
	std::fill(local_sep_order+sep_begin, local_sep_order+sep_end, integer_t(0));
	integer_t parts = 0;
	split(tree, parts, 0, 1);
	for (integer_t part=0, count=sep_begin+sub_graph_range.first; part<parts; part++)
	  for (integer_t i=sep_begin; i<sep_end; i++)
	    if (local_sep_order[i] == part) local_sep_order[i] = -count++;
	for (integer_t i=sep_begin; i<sep_end; i++) local_sep_order[i] = -local_sep_order[i];
      }
      this->sep_HSS_tree[sep] = tree;
      if (this->left_child[sep] != -1) reorder_separator(this->left_child[sep]);
      if (this->right_child[sep] != -1) reorder_separator(this->right_child[sep]);
    };
    int dsep = 0; // find which leaf of the distributed separator tree is our local subgraph
    while (dsep < dist_nr_sep && !(proc_dist_sep[dsep] == mpi_rank(comm) && dist_lchild[dsep] == -1)) dsep++;
    if (this->nbsep > 0) reorder_separator(this->nbsep-1);
    my_sub_graph->clear_temp_data();
    for (integer_t i=0; i<n_local; i++) local_sep_iorder[local_sep_order[i]-sub_graph_range.first] = i;
    my_sub_graph->permute_local(local_sep_order, local_sep_iorder,
				sub_graph_range.first, sub_graph_range.second);
    auto rank = mpi_rank(comm);
    auto P = mpi_nprocs(comm);
    auto rcnts = new int[2*P];
    auto displs = rcnts + P;
    for (int p=0; p<P; p++) {
      rcnts[p] = sub_graph_ranges[p].second - sub_graph_ranges[p].first;
      displs[p] = sub_graph_ranges[p].first;
    }
    MPI_Allgatherv(local_sep_order, rcnts[rank], mpi_type<integer_t>(),
     		   global_sep_order, rcnts, displs, mpi_type<integer_t>(), comm);
    delete[] rcnts;
    delete[] local_sep_order;
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  MatrixReorderingMPI<scalar_t,real_t,integer_t>::distributed_separator_reordering(integer_t* global_sep_order) {
    auto rank = mpi_rank(comm);
    auto P = mpi_nprocs(comm);
    int dsep = 0;
    while (dsep < dist_nr_sep && !(proc_dist_sep[dsep] == rank && dist_lchild[dsep] != -1)) dsep++;
    auto dsep_begin = dist_sep_range.first;
    auto dsep_end = dist_sep_range.second;
    auto dim_dsep = dsep_end - dsep_begin;
    auto dsep_order = new integer_t[2*dim_dsep];
    auto dsep_iorder = dsep_order + dim_dsep;
    if (dim_dsep) {
      HSSPartitionTree tree(dim_dsep);
      if ((dim_dsep >= params::hss_minimum_front_size) &&
	  (dim_dsep > 2 * params::hss_minimum_partition_size)) {
	std::vector<integer_t> sep_csr_ptr, sep_csr_ind;
	std::function<void(HSSPartitionTree&, integer_t&, integer_t, integer_t)> split =
	  [&](HSSPartitionTree& hss_tree, integer_t& nr_parts, integer_t part, integer_t count) {
	  my_dist_sep->extract_separator_subgraph(dsep_begin, 0, dim_dsep, part, dsep_order,
						  sep_csr_ptr, sep_csr_ind);
	  idx_t ncon = 1, edge_cut = 0, two = 2, nvtxs = sep_csr_ptr.size()-1;
	  auto partitioning = new idx_t[dim_dsep];
	  int info = WRAPPER_METIS_PartGraphRecursive(&nvtxs, &ncon, sep_csr_ptr, sep_csr_ind,
						      &two, &edge_cut, partitioning);
	  if (info !=  METIS_OK) {
	    std::cerr << "METIS_PartGraphRecursive for separator reordering returned: " << info << std::endl;
	    exit(1);
	  }
	  hss_tree.c.resize(2);
	  for (integer_t i=0, j=0; i<dim_dsep; i++)
	    if (dsep_order[i] == part) {
	      auto p = partitioning[j++];
	      dsep_order[i] = -count - p;
	      hss_tree.c[p].size++;
	    }
	  for (integer_t p=0; p<2; p++)
	    if (hss_tree.c[p].size > 2 * params::hss_minimum_partition_size)
	      split(hss_tree.c[p], nr_parts, -count-p, count+2);
	    else std::replace(dsep_order, dsep_order+dim_dsep, -count-p, nr_parts++);
	  delete[] partitioning;
	};
	std::fill(dsep_order, dsep_order+dim_dsep, integer_t(0));
	integer_t parts = 0;
	split(tree, parts, 0, 1);
	my_dist_sep->clear_temp_data();
	for (integer_t part=0, count=dsep_begin; part<parts; part++)
	  for (integer_t i=0; i<dim_dsep; i++)
	    if (dsep_order[i] == part) dsep_order[i] = -count++;
	for (integer_t i=0; i<dim_dsep; i++) dsep_order[i] = -dsep_order[i];
      } else for (integer_t i=0; i<dim_dsep; i++) dsep_order[i] = i+dsep_begin;
      dsep_HSS_tree[dsep] = tree;
      for (integer_t i=0; i<dim_dsep; i++) dsep_iorder[dsep_order[i]-dsep_begin] = i;
    }
    auto rcnts = new int[2*P];
    auto displs = rcnts + P;
    for (int p=0; p<P; p++) {
      rcnts[p] = dist_sep_ranges[p].second - dist_sep_ranges[p].first;
      displs[p] = dist_sep_ranges[p].first;
    }
    MPI_Allgatherv(dsep_order, rcnts[rank], mpi_type<integer_t>(),
		   global_sep_order, rcnts, displs, mpi_type<integer_t>(), comm);
    delete[] rcnts;
    my_dist_sep->permute_rows_local_cols_global(global_sep_order, dsep_iorder, dsep_begin, dsep_end);
    delete[] dsep_order;
    my_sub_graph->permute_columns(global_sep_order);
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  MatrixReorderingMPI<scalar_t,real_t,integer_t>::get_local_graphs
  (CSRMatrixMPI<scalar_t,real_t,integer_t>* A_mpi) {
    int P = mpi_nprocs(comm);
    int rank = mpi_rank(comm);
    sub_graph_ranges.resize(P);
    dist_sep_ranges.resize(P);
    proc_dist_sep.resize(dist_nr_sep);
    for (size_t sep=0, p_local=0, p_dist=0; sep<dist_lchild.size(); sep++) {
      if (dist_lchild[sep] == -1) { // sep is a leaf, so it is the local graph of proces p
	sub_graph_ranges[p_local] = std::make_pair(dist_sep_ptr[sep], dist_sep_ptr[sep+1]);
	proc_dist_sep[sep] = p_local++;
      } else { // sep was computed using distributed nested dissection, assign it to process p_dist
	dist_sep_ranges[p_dist] = std::make_pair(dist_sep_ptr[sep], dist_sep_ptr[sep+1]);
	proc_dist_sep[sep] = p_dist++;
      }
    }
    sub_graph_range = sub_graph_ranges[rank];
    dist_sep_range = dist_sep_ranges[rank];
    my_sub_graph = A_mpi->get_sub_graph(this->order, sub_graph_ranges);
    my_dist_sep = A_mpi->get_sub_graph(this->order, dist_sep_ranges);
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  MatrixReorderingMPI<scalar_t,real_t,integer_t>::build_sep_tree_from_order
  (CSRMatrixMPI<scalar_t,real_t,integer_t>* A_mpi) {
    auto P = mpi_nprocs(comm);
    auto rank = mpi_rank(comm);
    auto n = A_mpi->size();
    auto sub_n = my_sub_graph->size();
    std::vector<integer_t> sub_graph_etree = sp_symetree_dist
      (my_sub_graph->get_ptr(), my_sub_graph->get_ptr()+1, my_sub_graph->get_ind(),
       sub_n, sub_graph_range.first);
    std::vector<integer_t> post_order = etree_postorder(sub_graph_etree);
    auto iwork = new integer_t[sub_n];
    for (integer_t i=0; i<sub_n; ++i) iwork[post_order[i]] = post_order[sub_graph_etree[i]];
    for (integer_t i=0; i<sub_n; ++i) sub_graph_etree[i] = iwork[i];
    this->nbsep = etree_to_separator_tree
      (sub_graph_etree, this->parent, this->left_child, this->right_child, this->level, this->sep_ptr);
    for (integer_t i=0; i<sub_n; i++) {
      iwork[post_order[i]] = i;
      post_order[i] += sub_graph_range.first;
    }
    my_sub_graph->permute_local(post_order.data(), iwork,
				sub_graph_range.first, sub_graph_range.second);
    delete[] iwork;

    auto global_post_order = new integer_t[n];
    for (integer_t i=0; i<n; i++) global_post_order[i] = i;
    auto rcnts = new int[2*P];
    auto displs = rcnts + P;
    for (int p=0; p<P; p++) {
      rcnts[p] = sub_graph_ranges[p].second - sub_graph_ranges[p].first;
      displs[p] = sub_graph_ranges[p].first;
    }
    MPI_Allgatherv(post_order.data(), rcnts[rank], mpi_type<integer_t>(), global_post_order,
     		   rcnts, displs, mpi_type<integer_t>(), comm);
    delete[] rcnts;

    this->iorder.resize(n);
    for (integer_t i=0; i<n; i++) this->iorder[this->order[i]] = i;
    for (integer_t i=0; i<n; i++) this->order[global_post_order[i]] = this->iorder[i];
    delete[] global_post_order;
    for (integer_t i=0; i<n; i++) this->iorder[this->order[i]] = i;
    std::swap(this->order, this->iorder);
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  MatrixReorderingMPI<scalar_t,real_t,integer_t>::allgather_local_order
  (CSRMatrixMPI<scalar_t,real_t,integer_t>* A_mpi, integer_t* local_order) {
    auto P = mpi_nprocs(comm);
    auto n = A_mpi->size();
    auto local_rows = A_mpi->local_rows();
    this->order.resize(n);
    auto rcnts = new int[2*P];
    auto displs = rcnts + P;
    for (int p=0; p<P; p++) {
      rcnts[p] = A_mpi->get_dist()[p+1] - A_mpi->get_dist()[p];
      displs[p] = A_mpi->get_dist()[p]; // need to copy because displs needs to be 'int'
    }
    MPI_Allgatherv(local_order, local_rows, mpi_type<integer_t>(), this->order.data(),
		   rcnts, displs, mpi_type<integer_t>(), comm);
    delete[] rcnts;
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  MatrixReorderingMPI<scalar_t,real_t,integer_t>::clear_tree_data() {
    MatrixReordering<scalar_t,real_t,integer_t>::clear_tree_data();
    dist_nr_sep = 0;
    std::vector<integer_t>().swap(dist_sep_ptr);
    std::vector<integer_t>().swap(dist_parent);
    std::vector<integer_t>().swap(dist_lchild);
    std::vector<integer_t>().swap(dist_rchild);
    std::vector<integer_t>().swap(dist_level);
    std::vector<integer_t>().swap(dist_sizes);
    std::vector<integer_t>().swap(proc_dist_sep);
    std::unordered_map<integer_t,HSSPartitionTree>().swap(dsep_HSS_tree);
    std::vector<std::pair<integer_t,integer_t>>().swap(dist_sep_ranges);
    std::vector<std::pair<integer_t,integer_t>>().swap(sub_graph_ranges);
    delete my_sub_graph; my_sub_graph = NULL;
    delete my_dist_sep;  my_dist_sep = NULL;
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  MatrixReorderingMPI<scalar_t,real_t,integer_t>::nested_dissection_print
  (integer_t n, integer_t nnz, int stratnum, int stratpar, bool verbose) {
    integer_t total_separators = 0;
    MPI_Allreduce(&this->nbsep, &total_separators, 1, mpi_type<integer_t>(), MPI_SUM, comm);
    total_separators += std::max(integer_t(0), dist_nr_sep - mpi_nprocs(comm));
    integer_t max_level = this->level.empty() ? 0 : *std::max_element(this->level.begin(), this->level.end());
    MPI_Allreduce(MPI_IN_PLACE, &max_level, 1, mpi_type<integer_t>(), MPI_MAX, comm);
    if (verbose) {
      std::cout << "# initial matrix:" << std::endl;
      std::cout << "#   - number of unknowns = " << n << std::endl;
      std::cout << "#   - number of nonzeros = " << nnz << std::endl;
      std::cout << "# nested dissection reordering:" << std::endl;
      std::cout << "#   - " << this->get_name() << " reordering" << std::endl;
      std::cout << "#   - strategy number = " << stratnum << std::endl;
      std::cout << "#   - strategy parameter = " << stratpar << std::endl;
      std::cout << "#   - number of separators = " << total_separators << std::endl;
      std::cout << "#   - number of levels = " << max_level << std::endl;
    }
  }

} // end namespace strumpack

#endif // MATRIX_REORDERING_MPI_HPP
