/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Pieter Ghysels, Francois-Henry Rouet, Xiaoye S. Li.
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */
#ifndef METIS_REORDERING_HPP
#define METIS_REORDERING_HPP

#include <functional>
#include <typeinfo>
#include "MatrixReordering.hpp"

namespace strumpack {

  template<typename scalar_t,typename real_t,typename integer_t>
  class MetisReordering : public MatrixReordering<scalar_t,real_t,integer_t> {
  public:
    MetisReordering();
    ~MetisReordering();
    std::string get_name() { return std::string("Metis"); }
    virtual int nested_dissection(CompressedSparseMatrix<scalar_t,real_t,integer_t>* A, bool diag,
				  int stratnum, int stratpar, bool verbose,
				  int nx, int ny, int nz);

  protected:
    integer_t _n;
    integer_t *_jc;
    integer_t *_ir;

    std::size_t _nodes;
    std::size_t _separators;
    std::vector<idx_t> _sizes;

    virtual void compute_ordering();
    void compute_parent();
    void compute_level();
    void compute_children();
    void compute_offset();
  };

  template<typename scalar_t,typename real_t,typename integer_t>
  MetisReordering<scalar_t,real_t,integer_t>::MetisReordering() : _jc(NULL), _ir(NULL) {}

  template<typename scalar_t,typename real_t,typename integer_t>
  MetisReordering<scalar_t,real_t,integer_t>::~MetisReordering() {}

  template<typename scalar_t,typename real_t,typename integer_t> int
  MetisReordering<scalar_t,real_t,integer_t>::nested_dissection
  (CompressedSparseMatrix<scalar_t,real_t,integer_t>* A, bool diag, int stratnum,
   int stratpar, bool verbose, int nx, int ny, int nz) {
    _n = A->size();
    _jc = A->get_ptr();
    _ir = A->get_ind();

    if (stratpar <= 0) {
      if (mpi_root())
	std::cerr << "# WARNING: nested dissection parameter is " << stratpar
		  << ", but should be > 0, setting to 8" << std::endl;
      stratpar = 8;
    }
    _nodes = std::max(integer_t(3), ( _n / stratpar ) / 2 * 2 + 1);
    _separators = _nodes / 2;
    _sizes.resize(_nodes + 1);

    this->nbsep = (integer_t)_nodes;
    this->order.resize(_n);
    this->iorder.resize(_n);
    this->parent.resize(_nodes);
    this->level.resize(_nodes);
    this->left_child.resize(_nodes);
    this->right_child.resize(_nodes);
    this->sep_ptr.resize(_nodes+1);

    compute_ordering();
    compute_parent();
    compute_level();
    compute_children();
    compute_offset();
    this->nested_dissection_print(_n, _jc[_n]-_jc[0], stratnum, stratpar, verbose);
    std::vector<idx_t>().swap(_sizes); // clear _sizes
    return 0;
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  MetisReordering<scalar_t,real_t,integer_t>::compute_ordering() {
    auto xadj = new idx_t[_n+1 + _jc[_n]];
    auto adjncy = xadj + _n+1;
    integer_t e = 0;
    for (integer_t j=0; j<_n; j++) {
      xadj[j] = e;
      for (integer_t t=_jc[j]; t<_jc[j+1]; t++)
	if (_ir[t] != j) adjncy[e++] = _ir[t];
    }
    xadj[_n] = e;
    if (e==0) if (mpi_root()) std::cerr << "# WARNING: matrix seems to be diagonal!" << std::endl;

    if (std::is_same<idx_t,integer_t>())
      METIS_NodeNDP(_n, xadj, adjncy, NULL, _nodes / 2 + 1, NULL,
		    reinterpret_cast<idx_t*>(this->iorder.data()),
		    reinterpret_cast<idx_t*>(this->order.data()), _sizes.data());
    else {
      auto iord = new idx_t[this->iorder.size() + this->order.size()];
      auto ord = iord + this->iorder.size();
      METIS_NodeNDP(static_cast<idx_t>(_n), xadj, adjncy, NULL, _nodes / 2 + 1, NULL,
		    iord, ord, _sizes.data());
      this->iorder.assign(iord, iord+this->iorder.size());
      this->order.assign(ord, ord+this->order.size());
      delete[] iord;
    }
    delete[] xadj;
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  MetisReordering<scalar_t,real_t,integer_t>::compute_parent() {
    // Generates the parent positions of all the nodes in a
    // subtree, except for its root, and returns the number
    // of nodes in the subtree. The subtree is traversed in
    // left-to-right postorder and the parent positions are
    // written into parent starting from position pos.
    // The root of the subtree is specified by v, its
    // position in a top-down level-by-level right-to-left
    // traversal of the tree.
    std::function<std::size_t(std::size_t, std::size_t)>
      parent = [&](std::size_t pos, std::size_t v) -> std::size_t {
      std::size_t s0;
      std::size_t s1;
      if (v < _separators) {
	std::size_t root;
	s0 = parent(pos, 2 * v + 2);
	s1 = parent(pos + s0, 2 * v + 1);
	root = pos + s0 + s1;
	this->parent[pos + s0 - 1] = root;
	this->parent[pos + s0 + s1 - 1] = root;
      }
      else
	std::tie(s0, s1) = std::make_tuple(0, 0);
      return s0 + s1 + 1;
    };
    parent(0, 0);
    this->parent[_nodes - 1] = -1;
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  MetisReordering<scalar_t,real_t,integer_t>::compute_level() {
    // Returns the height of the subtree rooted at v.
    std::function<std::size_t(std::size_t)> height = [&](std::size_t v) -> std::size_t {
      return v < _separators ? 1 + std::max(height(2 * v + 2), height(2 * v + 1)) :  0;
    };

    // Generates the levels of the nodes of a subtree and
    // returns the number of nodes in that subtree. Levels
    // are numbered bottom-up starting at 1. The argument
    // ell specifies the level of the root.
    std::function<std::size_t(std::size_t, std::size_t, std::size_t)>
      level = [&](std::size_t pos, std::size_t ell, std::size_t v) -> std::size_t {
      std::size_t s0;
      std::size_t s1;
      if (v < _separators) {
	s0 = level(pos, ell - 1, 2 * v + 2);
	s1 = level(pos + s0, ell - 1, 2 * v + 1);
      }
      else
	std::tie(s0, s1) = std::make_tuple(0, 0);
      this->level[pos + s0 + s1] = ell;
      return s0 + s1 + 1;
    };
    level(0, height(0) + 1, 0);
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  MetisReordering<scalar_t,real_t,integer_t>::compute_children() {
    // Generates the children positions of all the nodes in
    // a subtree and returns the number of nodes in that
    // subtree. Leaves are assigned position -1.
    std::function<std::size_t(std::size_t, std::size_t)>
      children = [&](std::size_t pos, std::size_t v) -> std::size_t {
      std::size_t s0;
      std::size_t s1;
      if (v < _separators) {
	std::size_t root;
	s0 = children(pos, 2 * v + 2);
	s1 = children(pos + s0, 2 * v + 1);
	root = pos + s0 + s1;
	this->left_child[root] = pos + s0 - 1;
	this->right_child[root] = pos + s0 + s1 - 1;
      }
      else {
	std::tie(s0, s1) = std::make_tuple(0, 0);
	this->left_child[pos] = -1;
	this->right_child[pos] = -1;
      }
      return s0 + s1 + 1;
    };
    children(0, 0);
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  MetisReordering<scalar_t,real_t,integer_t>::compute_offset() {
    // Generates the offsets of the nodes of a subtree in
    // the graph ordering and returns the numbers of nodes
    // and vertices in the subtree. The argument offs
    // indicates the offset of the subtree in the graph
    // ordering.
    using ss = std::pair<std::size_t, std::size_t>;
    std::function<ss(std::size_t, std::size_t, std::size_t)>
      offset = [&](std::size_t pos, std::size_t offs, std::size_t v) -> ss {
      std::size_t s0;
      std::size_t s1;
      std::size_t q0;
      std::size_t q1;
      if (v < _separators) {
	std::tie(s0, q0) = offset(pos, offs, 2 * v + 2);
	std::tie(s1, q1) = offset(pos + s0, offs + q0, 2 * v + 1);
      }
      else {
	std::tie(s0, s1) = std::make_pair(0, 0);
	std::tie(q0, q1) = std::make_pair(0, 0);
      }
      this->sep_ptr[pos + s0 + s1] = offs + q0 + q1;
      return std::make_pair(s0 + s1 + 1, q0 + q1 + _sizes[_nodes - (v + 1)]);
    };

    std::size_t vertices;
    std::tie(std::ignore, vertices) = offset(0, 0, 0);
    this->sep_ptr[_nodes] = vertices;
  }

} // end namespace strumpack

#endif
