/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Pieter Ghysels, Francois-Henry Rouet, Xiaoye S. Li.
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */
#ifndef PARMETIS_REORDERING_HPP
#define PARMETIS_REORDERING_HPP

#include <typeinfo>
#include "MetisReordering.hpp"
#include "mpi.h"
#include "parmetis.h"

namespace strumpack {

  template<typename scalar_t,typename real_t,typename integer_t>
  class ParMetisReordering : public MatrixReorderingMPI<scalar_t,real_t,integer_t> {
  public:
    ParMetisReordering(MPI_Comm c) : MatrixReorderingMPI<scalar_t,real_t,integer_t>(c) {}
    ~ParMetisReordering() {}
    std::string get_name() { return std::string("ParMetis"); }
    int nested_dissection(CompressedSparseMatrix<scalar_t,real_t,integer_t>* A,
			  bool diag, int stratnum, int stratpar, bool verbose,
			  int nx, int ny, int nz);
  };

  template<typename integer_t> inline int WRAPPER_ParMETIS_V32_NodeND
  (std::vector<integer_t>& dist, idx_t* xadj, idx_t* adjncy, idx_t numflag,
   integer_t* local_order, idx_t* sizes, MPI_Comm c, integer_t local_rows) {
    idx_t* vtxdist = new idx_t[dist.size()+1+local_rows];
    idx_t* loc_order = vtxdist + dist.size()+1;
    for (size_t i=0; i<dist.size(); i++) vtxdist[i] = dist[i];
    int ierr = ParMETIS_V32_NodeND(vtxdist, xadj, adjncy, NULL, &numflag, NULL, NULL, NULL,
				   NULL, NULL, NULL, NULL, loc_order, sizes, &c);
    for (integer_t i=0; i<local_rows; i++) local_order[i] = loc_order[i];
    delete[] vtxdist;
    return ierr;
  }
  template<> inline int WRAPPER_ParMETIS_V32_NodeND
  (std::vector<idx_t>& dist, idx_t* xadj, idx_t* adjncy, idx_t numflag,
   idx_t* local_order, idx_t* sizes, MPI_Comm c, idx_t local_rows) {
    return ParMETIS_V32_NodeND(dist.data(), xadj, adjncy, NULL, &numflag, NULL, NULL, NULL,
			       NULL, NULL, NULL, NULL, local_order, sizes, &c);
  }

  template<typename scalar_t,typename real_t,typename integer_t> int
  ParMetisReordering<scalar_t,real_t,integer_t>::nested_dissection
  (CompressedSparseMatrix<scalar_t,real_t,integer_t>* A, bool diag,
   int stratnum, int stratpar, bool verbose, int, int, int) {
    CSRMatrixMPI<scalar_t,real_t,integer_t>* A_mpi = dynamic_cast<CSRMatrixMPI<scalar_t,real_t,integer_t>*>(A);
    if (A_mpi == NULL) {
      std::cerr << "ParMETIS reordering only works on CSRMatrixMPI (for now)!" << std::endl;
      return 1;
    }
    auto P = mpi_nprocs(this->comm);
    auto local_rows = A_mpi->local_rows();
    auto n = A_mpi->size();
    auto ptr = A_mpi->get_ptr();
    auto ind = A_mpi->get_ind();

    auto xadj = new idx_t[local_rows+1];
    auto adjncy = new idx_t[ptr[local_rows] - ptr[0]];
    integer_t e = 0;
    xadj[0] = 0;
    for (integer_t i=0; i<local_rows; i++) { // remove diagonal elements
      for (integer_t j=ptr[i]-ptr[0]; j<ptr[i+1]-ptr[0]; j++)
	if (ind[j]!=i+A_mpi->begin_row()) adjncy[e++] = ind[j];
      xadj[i+1] = e;
    } // xadj will now point locally, adjncy still has global column/row indices

    idx_t numflag = 0;
    int p2 = std::pow(2, std::floor(std::log2(P)));
    auto local_order = new integer_t[local_rows];
    auto sizes = new idx_t[2*p2-1];
    int ierr = WRAPPER_ParMETIS_V32_NodeND(A_mpi->get_dist(), xadj, adjncy, numflag,
					   local_order, sizes, this->comm, local_rows);
    if (ierr != METIS_OK) {
      if (mpi_rank(this->comm)==0) std::cerr << "ParMETIS_V32_NodeND failed with ierr=" << ierr << std::endl;
      return ierr;
    }
    delete[] xadj;
    delete[] adjncy;

    this->allgather_local_order(A_mpi, local_order);
    delete[] local_order;

    this->dist_nr_sep = 2*p2 - 1; // p2 subdomains, p2-1 distributed separators
    this->dist_lchild.assign(this->dist_nr_sep, integer_t(-1));
    this->dist_rchild.assign(this->dist_nr_sep, integer_t(-1));
    this->dist_parent.assign(this->dist_nr_sep, integer_t(-1));
    this->dist_sep_ptr.resize(this->dist_nr_sep+1);
    this->dist_level.resize(this->dist_nr_sep);
    int nr_dist_levels = std::log2(p2);
    std::function<void(integer_t,integer_t&,integer_t)> build_dist_binary_separator_tree =
      [&](integer_t dsep, integer_t& pid, integer_t level) {
      if (level > 0) {
	build_dist_binary_separator_tree(2*dsep+2, pid, level-1);
	integer_t lch = pid - 1;
	build_dist_binary_separator_tree(2*dsep+1, pid, level-1);
	this->dist_lchild[pid] = lch;
	this->dist_rchild[pid] = pid-1;
	this->dist_parent[lch] = pid;
	this->dist_parent[pid-1] = pid;
      }
      this->dist_level[pid] = nr_dist_levels - level;
      this->dist_sep_ptr[pid+1] = sizes[this->dist_nr_sep-1-dsep] + this->dist_sep_ptr[pid];
      pid++;
    };
    this->dist_sep_ptr[0] = 0;
    integer_t pid = 0;
    build_dist_binary_separator_tree(0, pid, nr_dist_levels);
    delete[] sizes;

    this->get_local_graphs(A_mpi);
    this->build_sep_tree_from_order(A_mpi);
    this->nested_dissection_print(n, A->nnz(), stratnum, stratpar, verbose);
    return 0;
  }

} // end namespace strumpack

#endif
