/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Pieter Ghysels, Francois-Henry Rouet, Xiaoye S. Li.
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 */
/*! \file strumpack_parameters.hpp
 * \brief Contains some global parameters, functions and enumerations.
 */
#ifndef STRUMPACK_PARAMETERS_HPP
#define STRUMPACK_PARAMETERS_HPP
#include <atomic>
#include <string>
#include <cmath>
#include <iostream>
#include <fstream>

namespace strumpack { // these are all global variables
  namespace params {
    /*!
     * \defgroup Enumerations Enumerations
     *           Public enumeration types. Enumerations for the C interface are not listed here.
     */

    /*! \brief Nested dissection code.
     * \ingroup Enumerations */
    enum MatrixReorderingStrategy {
      METIS,      /*!< Use either Metis (StrumpackSparseSolver, StrumpackSparseSolverMPI) or ParMetis (StrumpackSparseSolverMPI)    */
      SCOTCH,     /*!< Use either Scotch (StrumpackSparseSolver, StrumpackSparseSolverMPI) or PT-Scotch (StrumpackSparseSolverMPI)  */
      GEOMETRIC   /*!< A simple geometric nested dissection code that only works for regular meshes. (see Sp::reorder)              */
    };

    /*! \brief Type of Gram-Schmidt orthogonalization used in GMRes.
     * \ingroup Enumerations */
    enum GramSchmidtType {
      CLASSICAL,   /*!< Classical Gram-Schmidt is faster, more scalable.   */
      MODIFIED     /*!< Modified Gram-Schmidt is slower, but stable.       */
    };

    /*! \brief The random number distribution.
     * \ingroup Enumerations */
    enum RandomDistribution {
      NORMAL,   /*!< Normal(0,1) distributed numbers (takes roughly 23 flops per random number).  */
      UNIFORM   /*!< Uniform [0,1] distributed numbers (takes about 7 flops per random number).   */
    };

    /*! \brief Random number engine.
     * \ingroup Enumerations */
    enum RandomEngine {
      LINEAR,   /*!< The C++11 std::minstd_rand random number generator. */
      MERSENNE  /*!< The C++11 std::mt19937 random number generator.     */
    };

    /*! \brief Type of outer iterative (Krylov) solver.
     * \ingroup Enumerations */
    enum KrylovSolver {
      AUTO,           /*!< Use iterative refinement if no HSS compression is used, otherwise use GMRes.      */
      DIRECT,         /*!< No outer iterative solver, just a single application of the multifrontal solver.  */
      REFINE,         /*!< Iterative refinement.                                                             */
      PREC_GMRES,     /*!< Preconditioned GMRes. The preconditioner is the (approx) multifrontal solver.     */
      GMRES,          /*!< UN-preconditioned GMRes. (for testing mainly)                                     */
      PREC_BICGSTAB,  /*!< Preconditioned BiCGStab. The preconditioner is the (approx) multifrontal solver.  */
      BICGSTAB        /*!< UN-preconditioned BiCGStab. (for testing mainly)                                  */
    };

    /*! \brief Enumeration for the possible return codes.
     * \ingroup Enumerations */
    enum returnCode {
      SUCCESS=0,        /*!< Operation completed successfully. */
      MATRIX_NOT_SET,   /*!< The input matrix was not set.     */
      REORDERING_ERROR  /*!< The matrix reordering failed.     */
    };

    /*! \brief Strategy to dermine the approximate rank, or the number of random vectors.
     * \ingroup Enumerations */
    enum RankPattern {
      ADAPTIVE,       /*!< Use adaptive rank determination, increase number of random vectors until either rctol or actol is reached.
			   Not supported in the MPI code.                                                                             */
      CONSTANT,       /*!< Use a constant (params::random_vector_blocksize) number of random vectors.                                 */
      SQRTN,          /*!< Set number of random vectors to sqrt(dim_sep) + rank_offset.                                               */
      SQRTNLOGN,      /*!< Set number of random vectors to sqrt(dim_sep)*log(dim_sep) + rank_offset.                                  */
      BISECTIONCUT    /*!< Not supported yet.                                                                                         */
    };

#ifndef DOXYGEN_SHOULD_SKIP_THIS
    extern int separator_ordering_level; // not sure what this is, leave at 1 for now
    extern int hss_minimum_front_size;
    extern int hss_minimum_partition_size;
    extern int hss_coarsening_factor;  // please leave this at 2
    extern int random_vector_blocksize;
    extern int task_recursion_cutoff_level;
    extern int rank_offset;
    extern float rank_factor;
    extern int maximum_rank;
    extern int mc64job;
    extern int q_power;
    extern bool log_etree;
    extern bool log_ranks;
    // TODO make this a static constant
    extern int random_vector_panel;      // this is used to split the number of random vectors in blocks for tasking

    extern RandomDistribution random_distribution;
    extern RandomEngine random_engine;

    template<typename real> inline real get_default_relative_compression_tolerance() { return real(1.e-2); }
    template<typename real> inline real get_default_absolute_compression_tolerance() { return real(1.e-10); }
    template<typename real> inline real get_default_Krylov_relative_tolerance() { return real(1.e-6); }
    template<typename real> inline real get_default_Krylov_absolute_tolerance() { return real(1.e-10); }

    template<> inline float get_default_relative_compression_tolerance<float>() { return 1.e-1; }
    template<> inline float get_default_absolute_compression_tolerance<float>() { return 1.e-6; }
    template<> inline float get_default_Krylov_relative_tolerance<float>() { return 1.e-4; }
    template<> inline float get_default_Krylov_absolute_tolerance<float>() { return 1.e-6; }

    extern int num_threads;

    extern long long int total_mem;
    extern long long int max_mem;
    extern int mem_new_free;

    extern std::atomic<long long int> flops;
    extern std::atomic<long long int> bytes;
#endif /* DOXYGEN_SHOULD_SKIP_THIS */
  }

#ifndef DOXYGEN_SHOULD_SKIP_THIS
  /*! \brief Hold several options regarding the HSS preconditioning
   *         strategy.
   */
  template<typename real_t> class HSSStrategy {
  public:
    bool use_hss;
    real_t rctol;
    real_t actol;
    params::RankPattern rank_pattern;
    HSSStrategy() : use_hss(false),
		    rctol(params::get_default_relative_compression_tolerance<real_t>()),
		    actol(params::get_default_absolute_compression_tolerance<real_t>()),
		    rank_pattern(params::SQRTNLOGN) {}
    std::string get_rank_pattern() {
      switch (rank_pattern) {
      case params::ADAPTIVE:     return "ADAPTIVE";     break;
      case params::CONSTANT:     return "CONSTANT";     break;
      case params::SQRTN:        return "SQRTN";        break;
      case params::SQRTNLOGN:    return "SQRTNLOGN";    break;
      case params::BISECTIONCUT: return "BISECTIONCUT"; break;
      }
      return "";
    }
    int get_nr_random_vectors(int separator_size) {
      int nr_random_vectors;
      switch (rank_pattern) {
      case params::BISECTIONCUT: std::cerr << "ERROR: Rank pattern " << rank_pattern << " not supported." << std::endl;
      case params::SQRTN: {
	int mesh_size = std::sqrt(separator_size);
	nr_random_vectors = params::rank_factor * mesh_size + params::rank_offset;
      } break;
      case params::SQRTNLOGN: {
	int mesh_size = std::sqrt(separator_size);
	nr_random_vectors = params::rank_factor * mesh_size * std::log(mesh_size) + params::rank_offset;
      } break;
      case params::CONSTANT:
      case params::ADAPTIVE:
      default: {
	nr_random_vectors = params::random_vector_blocksize;
      } break;
      }
      return nr_random_vectors;
    }
  };
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

}

#ifndef DOXYGEN_SHOULD_SKIP_THIS
#ifdef COUNT_FLOPS
#define STRUMPACK_FLOPS(n) strumpack::params::flops += n;
#define STRUMPACK_BYTES(n) strumpack::params::bytes += n;
#else
#define STRUMPACK_FLOPS(n) void(0);
#define STRUMPACK_BYTES(n) void(0);
#endif
#endif /* DOXYGEN_SHOULD_SKIP_THIS */

#endif // STRUMPACK_PARAMETERS_HPP
